import re
import string

import mingus.containers as containers

def merge(a, b):
    '''
    (dict,dict) -> dict
    Combine two dictionaries without side-effects.
    '''
    result = dict(a)
    result.update(b)
    return result

def range_apply(func, list_list):
    return [[func(cell) for cell in column] for column in list_list]

def to_integer(scientific_note):
    '''
    :param scientific_note: Scientific representation of the note, like A4

    Returns an integer so you can feed this into spreadsheet functions
    '''
    return int(scientific_note)

def from_integer(integral_note):
    '''
    :param integral_note: Integral representation of a note, like 42

    Returns the note in scientific notation so you can feed this into
    other sheetmusic functions
    '''
    return containers.Note(integral_note)

def from_scientific(scientific_note):
    m = re.match(r'^([^0-9]+)([0-9]+)$', str(scientific_note))
    if not m:
        raise ValueError('%s is not in scientific notation' % scientific_note)
    note_within_octave = m.group(1)
    octave = int(m.group(2))
    return containers.Note(note_within_octave, octave)

def maybe_from_scientific(scientific_note):
    try:
        return from_scientific(scientific_note)
    except ValueError:
        return None

def to_scientific(integral_note):
    note = containers.Note(int(integral_note))
    return note.name + str(note.octave)

def parse_range_ref(Gnumeric, range_ref):
    workbook_index = 0 # Let's just hope that's always the case
    workbook = Gnumeric.workbooks()[workbook_index]
    sheet_index = int(Gnumeric.functions['sheet'](range_ref)) - 1
    sheet = workbook.sheets()[sheet_index]
    
    rows_list_or_float = Gnumeric.functions['row'](range_ref)
    if type(rows_list_or_float) == list:
        top = int(min(rows_list_or_float[0]))
        bottom = int(max(rows_list_or_float[0]))
    elif type(rows_list_or_float) == float:
        top = bottom = int(rows_list_or_float)
    
    columns_list_or_float = Gnumeric.functions['column'](range_ref)
    if type(columns_list_or_float) == list:
        left = int(min(x[0] for x in columns_list_or_float))
        right = int(max(x[0] for x in columns_list_or_float))
    elif type(columns_list_or_float) == float:
        left = right = int(columns_list_or_float)

    return left - 1, top - 1, right - 1, bottom - 1

def transpose(matrix):
    return zip(*matrix)

def cell_string_to_pos(cell_string):
    column = string.ascii_uppercase.index(re.match('^([A-Z]).*', cell_string, flags = re.IGNORECASE).group(1).upper())
    row = int(re.match(r'^[A-Z]+([0-9]+)$', cell_string, flags = re.IGNORECASE).group(1)) - 1
    return column, row

def parse_range_string(range_string):
    (left, top), (right, bottom) = map(cell_string_to_pos, range_string.split(':'))
    return left, top, right, bottom

def cell_positions(left, top, right, bottom):
    return [[(x,y) for y in range(top, bottom + 1)] for x in range(left, right + 1)]

def rendered_text(Gnumeric, x, y, workbook = 0, sheet = 1.0):
    sheet = Gnumeric.workbooks()[workbook].sheets()[int(sheet) - 1]
    return sheet.cell_fetch(x,y).get_rendered_text()

def italic(Gnumeric, x, y, workbook = 0, sheet = 1.0):
    sheet = Gnumeric.workbooks()[workbook].sheets()[int(sheet) - 1]
    result = sheet.cell_fetch(x,y).get_style().get_font_italic()
    return {0:False,1:True}[result]

def range_rendered_text(Gnumeric, left, top, right, bottom, workbook = 0, sheet = 1.0):
    columns = cell_positions(left, top, right, bottom)
    return [[rendered_text(Gnumeric, x,y) for x, y in column] for column in columns]

def range_rendered_text_and_italic(Gnumeric, left, top, right, bottom, workbook = 0, sheet = 1.0):
    columns = cell_positions(left, top, right, bottom)
    return [[(rendered_text(Gnumeric, x,y), italic(Gnumeric, x, y)) for x, y in column] for column in columns]

def iterate(func):
    'Decorate a function with this so I can pretend that I\'m using generators.'
    TIMES = 100
    def wrapper(*args, **kwargs):
        out = []
        iterable = iter(func(*args, **kwargs))
        for _ in range(TIMES):
            try:
                out.append(next(iterable))
            except StopIteration:
                break
        return out
    return wrapper

def repeat(func):
    TIMES = 30
    def wrapper(*args, **kwargs):
        result = list(func(*args, **kwargs))
        out = []
        for column in result:
            out.append(column * TIMES)
        return out
    return wrapper



plugin_xml = '''<?xml version="1.0"?>
<plugin id="Gnumeric_JkFuncPlugin">
  <information>
    <name>Sheetmusic</name>
    <description>Make music from spreadsheets</description>
  </information>
  <loader type="Gnumeric_PythonLoader:python">
    <attribute name="module_name" value="sheetmusic"/>
  </loader>
  <services>
    <service type="function_group" id="sheetmusic">
      <category>Local Python</category>
      <functions>
      </functions>
    </service>
  </services>
</plugin>
'''

sheetmusic_py = '''import libsheetmusic.main
sheetmusic_functions = libsheetmusic.main.functions()
'''



import functools
import itertools

import libsheetmusic.spreadsheet as s
import libsheetmusic.util as u

def note_functions():
    method_names = ['octave_up', 'octave_down']
    template = 'lambda root: s.note_method("%s", root)'
    return {method_name: eval(template % method_name) for method_name in method_names}

def interval_functions():
    special = 'from_shorthand',
    maybe_useful = ['get_interval', 'measure', 'invert',]
    questions = ['is_consonant', 'is_dissonant', 'is_imperfect_consonant', 'is_perfect_consonant',]

    keyed = ['second', 'third', 'fourth','fifth', 'sixth', 'seventh', 'unison']
    not_keyed = [
        'augmented_unison',
        'major_fifth', 'major_fourth', 'major_second', 'major_seventh', 'major_sixth', 'major_third', 'major_unison',
        'minor_fifth', 'minor_fourth', 'minor_second', 'minor_seventh', 'minor_sixth', 'minor_third', 'minor_unison',
        'perfect_fifth', 'perfect_fourth', 'unison'
    ]
    keyed_template = 'lambda root, key: s.keyed_interval("%s", root, key)'
    nonkeyed_interval = 'lambda root: s.nonkeyed_interval("%s", root)'
    keyed_functions = {name + '_interval': eval(keyed_template % name) for name in keyed}
    nonkeyed_functions = {name + '_interval': eval(nonkeyed_interval % name) for name in not_keyed}
    named = u.merge(keyed_functions, nonkeyed_functions)
    fancy = {'interval': _interval}
    return u.merge(named, fancy)

def _interval(majorminor, number, root):
    words = { 1: 'unison', 2: 'second', 3: 'third', 4: 'fourth',
              5: 'fifth',  6: 'sixth',  7: 'seventh' }
    if number in words:
        return s.nonkeyed_interval('%s_%s' % (majorminor, words[number]), root)

def scale_functions():
    scale_names = [
        'aeolian',
        'chromatic',
        'diatonic',
        'diminished',
        'dorian',
        'harmonic_minor',
        'ionian',
        'locrian',
        'lydian',
        'melodic_minor',
        'mixolydian',
        'natural_minor',
        'phrygian',
        'whole_note'
    ]
    template = 'lambda base: s.scale("%s", base)'
    return {name + '_scale':  eval(template % name) for name in scale_names}

def chord_functions():
    chord_names = [
        'I', 'I7', 'II', 'II7', 'III', 'III7', 'IV', 'IV7', 'V', 'V7', 'VI', 'VI7', 'VII', 'VII7',
        'augmented_major_seventh', 'augmented_minor_seventh', 'augmented_triad',
        'diminished_seventh', 'diminished_triad',
        'dominant', 'dominant7', 'dominant_flat_five', 'dominant_flat_ninth', 'dominant_ninth',
        'dominant_seventh', 'dominant_sharp_ninth', 'dominant_sixth','dominant_thirteenth',
        'eleventh', 'first_inversion',
        'half_diminished_seventh', 'hendrix_chord',
    #   'ii', 'ii7', 'iii', 'iii7', 'vi', 'vi7', 'vii', 'vii7',
        'invert', 'lydian_dominant_seventh',
        'major_ninth', 'major_seventh', 'major_sixth', 'major_thirteenth', 'major_triad',
        'mediant', 'mediant7', 'minor_eleventh', 'minor_major_seventh', 'minor_ninth',
        'minor_seventh', 'minor_seventh_flat_five', 'minor_sixth', 'minor_thirteenth',
        'minor_triad', 'second_inversion', 'seventh', 'sixth_ninth',
        'subdominant', 'subdominant7', 'submediant', 'submediant7', 'subtonic', 'subtonic7',
        'supertonic', 'supertonic7',
        'suspended_fourth_ninth', 'suspended_fourth_triad', 'suspended_second_triad',
        'suspended_seventh', 'suspended_triad', 'third_inversion',
        'tonic', 'tonic7', 'triad',
    ]
    chord_template = 'lambda base: s.chord("%s", base)'
    arpeggio_template = 'lambda base: s.arpeggio("%s", base)'

    chord = {name + '_chord': eval(chord_template % name) for name in chord_names}
    arpeggio = {name + '_arpeggio': eval(arpeggio_template % name) for name in chord_names}
    return u.merge(chord, arpeggio)

def gnumeric_functions():
    try:
        import Gnumeric
    except ImportError:
        def f(*args, **kwargs):
            raise EnvironmentError('This must be run from inside Gnumeric.')
        rep = progression = f
    else:
        def progression(progression_range_ref, string_root_note):
            return s.progression(Gnumeric, progression_range_ref, string_root_note)
        def rep(range_ref, times):
            return s.rep(Gnumeric, range_ref, times)

    return {
        'progression': progression,
        'rep': rep,
    }

def util_functions():
    func_names = ['_'.join(xs) for xs in itertools.product(['from','to'],['scientific','integer'])]
    return {func_name: getattr(u, func_name) for func_name in func_names}

custom_functions = {}
def function_functions():
    def define(x):
        result = eval(x)
        custom_functions[x] = result
        return x

    def call(x, *args):
        return custom_functions[x](*args)

    return {'define': define, 'call': call}

def functions():
    fs = [
        note_functions, scale_functions, chord_functions,
        interval_functions, util_functions,
        gnumeric_functions, function_functions
    ]
    return reduce(u.merge, [f() for f in fs])



import re
from itertools import repeat

import mingus.containers as containers
import mingus.core.chords as chords
import mingus.core.scales as scales
import mingus.core.intervals as intervals
import mingus.core.progressions as progressions

def scale(func_name, note):
    'http://code.google.com/p/mingus/wiki/tutorialScales'
    return [_ascending(note, getattr(scales, func_name)(note.name))]

def chord(*args, **kwargs):
    return [[x] for x in _chord(*args, **kwargs)]

def arpeggio(*args, **kwargs):
    return [_chord(*args,**kwargs)]

def progression(the_progression, root_note):
    '''
    >>> progression(['I, IV, V'], 'C3')
    [['C3', 'E3', 'G3'], ['F3', 'A4', 'C4'], ['G3', 'B4', 'D4']]
    '''
    named_chords = progressions.to_chords(re.split(r', ?', the_progression), root_note.name)

    def convert(next_chord, prev = {'note':root_note}):
        result = _ascending(prev['note'], next_chord)
        prev['note'] = result[0]
        return result

    return zip(*[convert(chord) for chord in named_chords])

def nonkeyed_interval(func_name, note):
    other_note_name = getattr(intervals, func_name)(note.name)
    return _next_note(note, other_note_name)

def keyed_interval(func_name, note, key):
    other_note_name = getattr(intervals, func_name)(note.name, key)
    return _next_note(note, other_note_name)

def _next_note(first_note, second_note_name):
    '''
    mingus.containers.Note -> str -> mingus.containers.Note

    Receive a Note and a string note specifying a note name without
    an octave.

    If the second note name is greater than or equal to the first
    ("G" is greater than "C"),
    return a note of the second name in the same octave as the first.

    If the second note name is less than the first ("G" is greater than "C"),
    return a note of the second name in the octave above the first.
    '''
    if second_note_name >= first_note.name:
        second_note_octave = first_note.octave
    else:
        second_note_octave = first_note.octave + 1
    return containers.Note(second_note_name, second_note_octave)

def _chord(func_name, note, *args, **kwargs):
    'http://code.google.com/p/mingus/wiki/tutorialChords'
    return _ascending(note, getattr(chords, func_name)(note.name, *args, **kwargs))

def _ascending(note, note_names):
    prev_note = note
    output = []
    for note_name in note_names:
        next_note = _next_note(prev_note, note_name)
        output.append(next_note)
        prev_note = next_note
    return output



import functools
from itertools import izip
from Queue import Queue
from threading import Thread

import mingus.containers as c
import mingus.extra.LilyPond as LilyPond
from mingus.midi import fluidsynth, MidiFileOut

import libsheetmusic.util as u

class Sub:
    def __init__(self, function):
        self.queue = Queue()
        def worker(queue):
            while queue.empty():
                function()
        self.thread = Thread(None, target = worker, args = (self.queue,))
        self.thread.start()
    def stop(self):
        self.queue.put(9839232)

def to_track(range_values, key, upper, lower):
    '''
    Convert the cells to sheet music.
    >>> to_track([['C3', 'C3'], ['E3', 'E3'], ['G3', 'G3'], ['C4', 'G4']], 'C', 2, 4)
    '''
    def note_and_italic(cell):
        maybe_note, italic = cell
        return u.maybe_from_scientific(maybe_note), italic
    cells = u.transpose(u.range_apply(note_and_italic, range_values))
    meter = (int(upper), int(lower))

    ncs = []
    for row in cells:
        nc = c.NoteContainer()
        for note, _ in row:
            if note != None:
                nc.add_note(note)
        allitalic = all(note_italic[1] for note_italic in row)
        ncs.append((nc, allitalic))

    t = c.Track()
    while len(ncs) > 0:
        nc, allitalic = ncs.pop(0)
        duration = 0.25
        while allitalic and len(ncs) > 0 and ncs[0][0] == nc:
            duration += 0.25
            _, allitalic = ncs.pop(0)
        if not t.add_notes(nc, duration = int(1/duration)):
            raise AssertionError('I failed to add a note container.')
    return t

def sheetmusic(Gnumeric, range_ref, key = "C", upper = 4, lower = 4):
    '''
    Convert the cells to sheet music.
    '''
    left, top, right, bottom = u.parse_range_ref(Gnumeric, range_ref)
    values = u.range_rendered_text(Gnumeric, left, top, right, bottom,
        sheet = 1) #Gnumeric.functions['sheet'](range_ref))
    t = to_track(values, key, upper, lower)
    lp = LilyPond.from_Track(t)
    return LilyPond.to_png(lp, '/tmp/track')

def midi(Gnumeric, fn, range_string, key = "C", upper = 4, lower = 4, bpm = 120):
    'Convert the cells to MIDI.'
    top, left, bottom, right = u.parse_range_string(range_string)
    MidiFileOut.write_Track(fn, to_track(u.rendered_text(Gnumeric, top, left, bottom, right), key, upper, lower), bpm = bpm)

def play(Gnumeric, range_string, key = "C", upper = 4, lower = 4, bpm = 120):
    'Play the music in some cells.'
    def track(range_string):
        top, left, bottom, right = u.parse_range_string(range_string)
        stuff = u.range_rendered_text_and_italic(Gnumeric, top, left, bottom, right)
        return to_track(stuff, key, upper, lower)
    fluidsynth.play_Track(track(range_string), bpm = bpm)

def loop(Gnumeric, range_string, key = "C", upper = 4, lower = 4, bpm = 120):
    'Loop the music in some cells.'
    return Sub(functools.partial(play, Gnumeric, range_string, key, upper, lower, bpm))

def init():
    sf = '/usr/share/soundfonts/Unison.sf2'
    fluidsynth.init(sf, 'alsa')
    fluidsynth.play_Note(c.Note('A'))

'''
def show_track(Gnumeric, range_string, key = "C", upper = 4, lower = 4, bpm = 120):
    'Play the music in some cells.'
    top, left, bottom, right = u.parse_range_string(range_string)
    track = to_track(u.rendered_text(Gnumeric, top, left, bottom, right), key, upper, lower)
    return track

def show_rendered_text(Gnumeric, range_string, key = "C", upper = 4, lower = 4, bpm = 120):
    'Play the music in some cells.'
    top, left, bottom, right = u.parse_range_string(range_string)
    return u.rendered_text(Gnumeric, top, left, bottom, right)
'''



from functools import partial

import libsheetmusic.console as c

try:
    import Gnumeric
except ImportError:
    pass
else:
    play = partial(c.play, Gnumeric)
    loop = partial(c.loop, Gnumeric)
    midi = partial(c.midi, Gnumeric)
    sheetmusic = partial(c.sheetmusic, Gnumeric)
    init = c.init
#   bold = partial(c.bold, Gnumeric)
#   show_track = partial(c.show_track, Gnumeric)
#   show_rendered_text = partial(c.show_rendered_text, Gnumeric)
#   show_range_coords = partial(c.show_range_coords, Gnumeric)



from itertools import repeat

import libsheetmusic.music as m
import libsheetmusic.util as u

@u.iterate
def scale(func_name, string_note):
    note_in = u.from_scientific(string_note)
    notes_out = m.scale(func_name, note_in)
    return u.range_apply(u.to_scientific, notes_out)

@u.iterate
def chord(func_name, string_note, *args, **kwargs):
    return u.range_apply(u.to_scientific, m.chord(func_name, u.from_scientific(string_note), *args, **kwargs))

@u.iterate
def arpeggio(func_name, string_note, *args, **kwargs):
    return u.range_apply(u.to_scientific, m.arpeggio(func_name, u.from_scientific(string_note), *args, **kwargs))

@u.repeat
def progression(Gnumeric, progression_range_ref, string_root_note):
    if 'RangeRef' in str(type(progression_range_ref)):
        left, top, right, bottom = u.parse_range_ref(Gnumeric, progression_range_ref)
        # the_progression = u.range_rendered_text(Gnumeric, progression_range_ref)
        raise NotImplementedError
    else:
        the_progression = progression_range_ref
    root_note = u.from_scientific(string_root_note)
    return u.range_apply(u.to_scientific, m.progression(the_progression, root_note))

def keyed_interval(func_name, string_note, key):
    key = str(key)[0]
    return u.to_scientific(m.keyed_interval(func_name, u.from_scientific(string_note), key))

def nonkeyed_interval(func_name, string_note):
    return u.to_scientific(m.nonkeyed_interval(func_name, u.from_scientific(string_note)))

def note_method(method_name, string_note):
    note = u.from_scientific(string_note)
    getattr(note, method_name)()
    return u.to_scientific(note)

@u.repeat
def rep(Gnumeric, range_ref, times):
    left, top, right, bottom = u.parse_range_ref(Gnumeric, range_ref)
    values = u.range_rendered_text(Gnumeric, left, top, right, bottom,
        sheet = 1) # Gnumeric.functions['sheet'](range_ref))
    for column in values:
        newcolumn = []
        for cell in column:
            newcolumn.extend([cell] * int(times))
        yield newcolumn



from distutils.core import setup

setup(name='sheetmusic',
      author='Thomas Levine',
      author_email='_@thomaslevine.com',
      description='Functions to be loaded by the Gnumeric Sheetmusic plugin',
      url='https://github.com/tlevine/sheetmusic',
      packages=['sheetmusic', 'libsheetmusic'],
      scripts=['bin/sheetmusic-installer'],
      install_requires = ['mingus', 'lxml'],
      tests_require = ['nose'],
      version='0.0.3',
      license='AGPL',
)



import os
import re
import subprocess

import lxml.etree

import libsheetmusic.main as main
import libsheetmusic.plugin_template as templates

def get_plugin_xml():
    plugin = lxml.etree.fromstring(templates.plugin_xml)
    functions = plugin.xpath('//functions')[0]
    for name in main.functions().keys():
        functions.insert(0, lxml.etree.Element('function', attrib = {'name': name}))
    return lxml.etree.tostring(plugin)

def get_version(system = os.system):
    proc = subprocess.Popen(['gnumeric', '--version'], stdout=subprocess.PIPE)
    out = proc.communicate()[0]
    version = re.match(r"^gnumeric version '([0-9.]+)'.*", out).group(1)
    return version

def install(os = os):
    directory = os.path.expanduser('~/.gnumeric/%s/plugins/sheetmusic' % get_version(os.system))
    try:
        os.makedirs(directory)
    except OSError:
        pass

    plugin_xml = os.path.join(directory, 'plugin.xml')
    sheetmusic_py = os.path.join(directory, 'sheetmusic.py')

    with open(plugin_xml, 'w') as fp:
        fp.write(get_plugin_xml())

    with open(sheetmusic_py, 'w') as fp:
        fp.write(templates.sheetmusic_py)
